/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.7 $
 * $Date: 2004/03/31 15:48:58 $
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "ArticleDownloader.h"
#include "Decoder.h"
#include "global.h"

#include "MultiServerPool.h"

ArticleDownloader::ArticleDownloader(ArticleJob* job, char* tempdir)
{
  this->job = job;
  this->job->setStatus(JOBSTATUS_ASSIGNED);

  if (tempdir)
    {
      this->tempdir = (char*) malloc(strlen(tempdir) + 2);
      memset(this->tempdir, 0, strlen(tempdir) + 2);
      memcpy(this->tempdir, tempdir, strlen(tempdir));
      if (this->tempdir[strlen(this->tempdir)-1] != '/')
        this->tempdir[strlen(this->tempdir)] = '/';
    }
  else
    this->tempdir = strdup("./");
  
}

ArticleDownloader::~ArticleDownloader()
{
  if (this->tempdir)
    free(this->tempdir);
}

void ArticleDownloader::run()
{
  this->job->setStatus(JOBSTATUS_RUNNING);
  int level = 0;
  while (1)
    {

      NNTPConnection* conn = MultiServerPool::getInstance()->getConnection(level);

      if (!conn)
        {
          // all connection-levels passed, still not finished...
          // seems like we failed.
          this->job->setStatus(JOBSTATUS_FAILED);
          error("%s FAILED! Article unavailable on all servers.", this->job->getMessageID());
          break;
        }

      // Okay, we got a connection. Now start downloading!!

      // but at first, change group! dryan's server wants it this way... ;-)
      bool grpchanged = true;

      for (unsigned int i = 0; i < this->job->getGroups()->size(); i++)
        {
          if (conn->joinGroup((*(this->job->getGroups()))[i]))
            {
              grpchanged = false;
            }
          else
            {
              grpchanged = true;
              break;
            }
        }

      if (grpchanged)
        {

          // now, let's begin!
          char tmp[1024];
      
          snprintf(tmp, 1024, "ARTICLE %s\r\n", this->job->getMessageID());

          char* answer = NULL;

          for (int retry = 3; retry > 0; retry--)
            {
              answer = conn->request(tmp);
              if (answer && (!strncmp(answer, "2", 1)))
                break;
            }

          if (answer && (!strncmp(answer, "2", 1)))
            {
              // positive answer!

              Frontend::getInstance()->started(this->job->getMessageID(), conn->getServer()->getHost());
              
              char* line = NULL;

              char tmpfilename[1024];

              snprintf(tmpfilename, 1024, "%snzbgettemp-XXXXXX", this->tempdir);

              int tmpfd = mkstemp(tmpfilename);

              FILE* outfile = fdopen(tmpfd, "w+");
              if (outfile)
                {

                  bool body = false;
                  int lines = 0;
                  int maxlines = 0;
                  int bytes = 0;

                  while (1)
                    {
                      line = conn->readLine();

                      //detect end of article
                      if ((!line) || (!strcmp(line, ".\r\n")) || (!strcmp(line, ".\n")))
                        break;

                      //detect lines starting with "." (marked as "..")
                      if (!strncmp(line, "..", 2))
                        line++;
                      
                      // detect Lines-header
                      if ((!maxlines) && (!strncmp(line, "Lines: ",7)))
                        maxlines = atoi(line+7);
                      
                      // count lines in body
                      if (body)
                        lines++;
                      
                      // detect end og header
                      if ((!body) && ((!strcmp(line,"\r\n")) || (!strcmp(line,"\n"))))
                        body = true;
                      
                      int wrcnt = fwrite(line, 1, strlen(line), outfile);
                      if (wrcnt > 0)
                        bytes += wrcnt;
                      
                      
                      Frontend::getInstance()->progress(this->job->getMessageID(),
                                                        conn->getServer()->getHost(),
                                                        bytes,
                                                        lines,
                                                        maxlines);
                    }
                  
                  fflush(outfile);
                  fclose(outfile);
                  
                  if (line)
                    {
                      Frontend::getInstance()->finished(this->job->getMessageID(), conn->getServer()->getHost());
                      MultiServerPool::getInstance()->freeConnection(level ,conn);
                      char* resFilename = Decoder::getInstance()->decode(tmpfilename, this->tempdir);
                      
                      if (resFilename)
                        {
                          info("Decoded %s", resFilename);
                          
                          this->job->setResultFilename(resFilename);
                          this->job->setStatus(JOBSTATUS_FINISHED);
                          
                          free(resFilename);
                          remove(tmpfilename);
                          break;
                        }
                      else 
                        {
                          error("Decoding %s failed", this->job->getMessageID());
                          // possibly permanent error, abort
                          this->job->setStatus(JOBSTATUS_FAILED);
                          break;
                        }
                    }
                  else
                    {
                      warn("Unexpected end of %s", this->job->getMessageID());
                    }

                  remove(tmpfilename);

                }
              else
                {
                  warn("Could not create file %s", tmpfilename);
                }
          
            }
          else
            {
              if (!answer)
                warn("%s @ %s failed: Connection closed by remote host.", this->job->getMessageID(), conn->getServer()->getHost());
              else
                warn("%s @ %s failed (%s)", this->job->getMessageID(), conn->getServer()->getHost(), answer);
              Frontend::getInstance()->finished(this->job->getMessageID(), conn->getServer()->getHost());
              
            }
        }
      else
        {
          //group-change failed. Ignore this, too noisy to report
        }
      

      MultiServerPool::getInstance()->freeConnection(level,conn);
      Frontend::getInstance()->finished(this->job->getMessageID(), conn->getServer()->getHost());
      
      level++;
    }

  this->stop();
}
