/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.9 $
 * $Date: 2004/03/30 12:21:03 $
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vector>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>

#include "FileDownloader.h"
#include "ArticleDownloader.h"
#include "ArticleJob.h"
#include "global.h"
#include "Observable.h"
#include "MultiServerPool.h"

FileDownloader::FileDownloader(DownloadJob* job, char* tempdir, char* destdir)
{
  this->job = job;
  this->childJobs.clear();
  this->children.clear();
  if (tempdir)
    {
      this->tempdir = (char*) malloc(strlen(tempdir) + 2);
      memset(this->tempdir, 0, strlen(tempdir) + 2);
      memcpy(this->tempdir, tempdir, strlen(tempdir));
      if (this->tempdir[strlen(this->tempdir)-1] != '/')
        this->tempdir[strlen(this->tempdir)] = '/';
    }
  else
    this->tempdir = strdup("./");

  if (destdir)
    {
      this->destdir = (char*) malloc(strlen(destdir) + 2);
      memset(this->destdir, 0, strlen(destdir) + 2);
      memcpy(this->destdir, destdir, strlen(destdir));
      if (this->destdir[strlen(this->destdir)-1] != '/')
        this->destdir[strlen(this->destdir)] = '/';
    }
  else
    this->destdir = strdup("./");
}

FileDownloader::~FileDownloader()
{
  if (this->tempdir)
    free(this->tempdir);
  
  if (this->destdir)
    free(this->destdir);

  for (unsigned int i = 0; i < this->childJobs.size(); i++)
    {
      if (this->childJobs[i])
        {
          delete this->childJobs[i];
          this->childJobs[i] = NULL;
        }
    }

  for (unsigned int i = 0; i < this->children.size(); i++)
    {
      if (this->children[i])
        {
          delete this->children[i];
          this->childJobs[i] = NULL;
        }
    }
}


void FileDownloader::start()
{

  if (this->dupe())
    {
      this->job->setStatus(JOBSTATUS_FINISHED);
      warn("File with subject \"%s\" seems to be duplicate, skipping.", 
           this->job->getSubject());
      return;
    }

  for (unsigned int i = 0; i < this->job->getMessageIDs()->size(); i++)
    {
      //start only when level-0-connection is available
      NNTPConnection* conn = MultiServerPool::getInstance()->getConnection(0);
      MultiServerPool::getInstance()->freeConnection(0, conn);

      ArticleJob* cjob = new ArticleJob((*(this->job->getMessageIDs()))[i]);

      for (unsigned int j = 0; j < this->job->getGroups()->size(); j++)
        {
          cjob->getGroups()->push_back((*(this->job->getGroups()))[j]);
        }

      cjob->addObserver(this);
      this->childJobs.push_back(cjob);
      
      ArticleDownloader* dl = new ArticleDownloader(cjob, this->tempdir);

      this->children.push_back(dl);
    }

  for (unsigned int i = 0; i < this->children.size(); i++)
    {
      //start only when free connections are available for every level
      int level = 0;
      while (1)
        {
          NNTPConnection* conn = MultiServerPool::getInstance()->getConnection(level);
          if (conn)
            MultiServerPool::getInstance()->freeConnection(level, conn);
          else
            break;
          level++;
        }

      this->children[i]->start();
      usleep(100*1000);
    }
}

void FileDownloader::notify(Observable* caller)
{
  debug("FileDownloader::notify(Observable* caller)");
  
  // all child-jobs ready?
  for (unsigned int i = 0; i < this->childJobs.size(); i++)
    {
      if ((this->childJobs[i]->getStatus() != JOBSTATUS_FINISHED)
          && (this->childJobs[i]->getStatus() != JOBSTATUS_FAILED))
        {
          return;
        }
    }

  // all ready!

  // determine filename

  char* basename = NULL;
  int maxpart = 0;
  
  for (unsigned int i = 0; i < this->childJobs.size(); i++)
    {
      char* partfn = this->childJobs[i]->getResultFilename();
      if (partfn)
        {
          char* lastdot = strrchr(partfn, '.');

          if (lastdot)
            {
              char* localbasename = (char*) malloc(lastdot-partfn+1);
              memset(localbasename, 0, lastdot-partfn+1);
              memcpy(localbasename, partfn, lastdot-partfn);

              if (strncmp(localbasename, "nzbgettemp-unknown-", 19))
                {
                  if (basename)
                    {
                      if (strcmp(basename, localbasename))
                        {
                          error("Filenames differ in parts of post!");
                          this->job->setStatus(JOBSTATUS_FAILED);
                          return;
                        }
                      else
                        free(localbasename);
                    }
                  else
                    {
                      basename = localbasename;
                    }
                }
              
              int localpart = atoi(lastdot+1);
              if (localpart > maxpart)
                maxpart = localpart;
            }
        }
    }

  if (!basename)
    {
      //ooh, something f*cked up, so let's create our own name
      basename = (char*) malloc(1024);
      snprintf(basename, 1024, "nzbget-unknown-XXXXXX");
      mktemp(basename);
    }

  // search childs for unnamed parts
  for (unsigned int i = 0; i < this->childJobs.size(); i++)
    {
      char* partfn = this->childJobs[i]->getResultFilename();
      if (partfn)
        {
          if (!strncmp(partfn, "nzbgettemp-unknown-", 19))
            {
              char oldfn[1024];
              char newfn[1024];
          
              snprintf(oldfn, 1024, "%s%s", this->tempdir, partfn);
              snprintf(newfn, 1024, "%s%s.%d", this->tempdir, basename, i+1);
          
              rename(oldfn, newfn);

              if ((signed int)i+1 > maxpart)
                maxpart = i+1;
            }
        }
    }

  FILE* outfile;

  char ofn[1024];
  snprintf(ofn, 1024, "%s%s",this->destdir, basename);


  // prevent overwriting existing files
  struct stat statbuf;
  int dupcount = 0;
  while (!stat(ofn, &statbuf))
    {
      dupcount++;
      snprintf(ofn, 1024, "%s%s_duplicate%d",this->destdir, basename, dupcount);
    }

  outfile = fopen(ofn, "w+");

  if (!outfile)
    {
      error("Could not create file %s!", basename);
      this->job->setStatus(JOBSTATUS_FAILED);
      return;
    }


  bool complete = true;

  for (int i = 1; i <= maxpart; i++)
    {
      FILE* infile;

      char fn[1024];
      snprintf(fn, 1024, "%s%s.%d", this->tempdir, basename, i);
      
      infile = fopen(fn, "r");

      if (infile)
        {
          char buffer[10240];

          int cnt = 10240;

          while (cnt == 10240)
            {
              cnt = fread(buffer, 1, 10240, infile);
              fwrite(buffer, 1, cnt, outfile);
            }
          
          fclose(infile);
          remove(fn);
        }
      else
        complete = false;
    }

  fclose(outfile);

  if (complete)
    {
      this->job->setStatus(JOBSTATUS_FINISHED);
      info("Decoded %s", ofn);
    }
  else
    {
      char brokenfn[1024];
      snprintf(brokenfn, 1024, "%s_broken", ofn);
      rename(ofn, brokenfn);
      this->job->setStatus(JOBSTATUS_FAILED);
      warn("%s is incomplete!", ofn);
    }
}


bool FileDownloader::isRunning()
{
  for (unsigned int i = 0; i < this->children.size(); i++)
    {
      if (this->children[i]->isRunning())
        return true;
    }
  return false;
}


bool FileDownloader::dupe()
{
  if ((!this->job) || (!this->job->getSubject()) || (!this->destdir))
    return false;

  DIR *dir=opendir(this->destdir);
  struct dirent *entry;
  if (!dir)
    return false;

  std::vector< char* > existing;

  existing.clear();
    
  while ((entry=readdir(dir))) 
    {
      if (!strcmp(entry->d_name, ".."))
        continue;

      if (!strcmp(entry->d_name,"."))
        continue;

      // do not consider short filenames, too many false positives!!
      if ((strlen(entry->d_name) < 6) && (strstr(entry->d_name, ".") == NULL))
        continue;

      existing.push_back(strdup(entry->d_name));
    }

  closedir(dir);

  // mmkay, got the names of existing files, now make substring-tests
  // with subject

  bool res = false;

  for (unsigned int i = 0; i < existing.size(); i++)
    {
      if (strstr(this->job->getSubject(), existing[i]))
        {
          // found a match!
          res = true;
        }
      free(existing[i]);
      existing[i] = NULL;
    }

  existing.clear();
  return res;
}
