/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.7 $
 * $Date: 2004/03/30 19:03:06 $
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>

#include "Frontend.h"

extern int threadCount;

Frontend::Frontend()
{
  this->messages.clear();
  this->running.clear();
  this->queueSize = 0;

  pthread_mutex_init(&mutex_running, NULL);
  pthread_mutex_init(&mutex_messages, NULL);
}

Frontend::~Frontend()
{
  
}

Frontend* Frontend::getInstance()
{
  static Frontend singleton;
  return &singleton;
}

void Frontend::started(char* messageid, char* server)
{
  runningDL* run = (runningDL*) malloc(sizeof(runningDL));
  run->messageid = messageid;
  run->server = server;
  run->bytes = 0;
  run->lines = 0;
  run->maxlines = 0;
  run->starttime = ::time(NULL);

  pthread_mutex_lock(&this->mutex_running);

  this->running.push_back(run);

  pthread_mutex_unlock(&this->mutex_running);
}
    
void Frontend::finished(char* messageid, char* server)
{
  pthread_mutex_lock(&this->mutex_running);

  for (unsigned int i=0; i<this->running.size(); i++)
    {
      if ((this->running[i]) && (this->running[i]->messageid == messageid) &&
          (this->running[i]->server == server))
        {
          runningDL* run = this->running[i];
          this->running.erase(this->running.begin()+i);
          free(run);
          break;
        }

    }

  pthread_mutex_unlock(&this->mutex_running);
}


void Frontend::progress(char* messageid, char* server, int bytes, int lines, int maxlines)
{
  pthread_mutex_lock(&this->mutex_running);

  for (unsigned int i=0; i<this->running.size(); i++)
    {
      if ((this->running[i]) && (this->running[i]->messageid == messageid) &&
          (this->running[i]->server == server))
        {
          this->running[i]->bytes = bytes;
          this->running[i]->lines = lines;
          this->running[i]->maxlines = maxlines;
          break;
        }
    }
  
  pthread_mutex_unlock(&this->mutex_running);
}

void Frontend::frontendError(char* msg, ...)
{
  char tmp[1024];
  char tmp2[1024];

  va_list ap;
  va_start(ap, msg);
  vsnprintf(tmp2, 1024, msg, ap);
  va_end(ap);


  snprintf(tmp, 1024, "\033[31m[ERROR]\033[39m %s\033[K\n", tmp2);

  pthread_mutex_lock(&this->mutex_messages);

  this->messages.push_back(strdup(tmp));

  pthread_mutex_unlock(&this->mutex_messages);
}

void Frontend::frontendWarn(char* msg, ...)
{
  char tmp[1024];
  char tmp2[1024];

  va_list ap;
  va_start(ap, msg);
  vsnprintf(tmp2, 1024, msg, ap);
  va_end(ap);

  snprintf(tmp, 1024, "\033[35m[WARNING]\033[39m %s\033[K\n", tmp2);

  pthread_mutex_lock(&this->mutex_messages);

  this->messages.push_back(strdup(tmp));

  pthread_mutex_unlock(&this->mutex_messages);
}

void Frontend::frontendDebug(char* msg, ...)
{
  char tmp[1024];
  char tmp2[1024];

  va_list ap;
  va_start(ap, msg);
  vsnprintf(tmp2, 1024, msg, ap);
  va_end(ap);

  snprintf(tmp, 1024, "[DEBUG] %s\033[K\n", tmp2);

  pthread_mutex_lock(&this->mutex_messages);

  this->messages.push_back(strdup(tmp));

  pthread_mutex_unlock(&this->mutex_messages);
}

void Frontend::frontendInfo(char* msg, ...)
{
  char tmp[1024];
  char tmp2[1024];

  va_list ap;
  va_start(ap, msg);
  vsnprintf(tmp2, 1024, msg, ap);
  va_end(ap);

  snprintf(tmp, 1024, "\033[32m[INFO]\033[39m %s\033[K\n", tmp2);

  pthread_mutex_lock(&this->mutex_messages);

  this->messages.push_back(strdup(tmp));

  pthread_mutex_unlock(&this->mutex_messages);
}

void Frontend::run()
{
  while(1)
    {
      this->printMessages();
      this->printStatus();

      usleep(500*1000);
    }
}

void Frontend::printMessages()
{
  pthread_mutex_lock(&this->mutex_messages);
  
  for (unsigned int i = 0; i < this->messages.size(); i++)
    {
      fprintf(stdout, "%s", this->messages[i]);
      free(this->messages[i]);
      this->messages[i] = NULL;
    }

  this->messages.clear();

  pthread_mutex_unlock(&this->mutex_messages);
}

void Frontend::printStatus()
{
  char tmp[1024];

  pthread_mutex_lock(&this->mutex_running);

  pthread_mutex_lock(&this->mutex_messages);

  int lines = 0;

  fprintf(stdout, "\033[s");

  int overall_speed = 0;

  for (unsigned int i = 0; i < this->running.size(); i++)
    {
      if (this->running[i])
        {
          
          int percent = 0;

          if (this->running[i]->maxlines > 0)
            percent = (100*this->running[i]->lines) / this->running[i]->maxlines;

          time_t tdiff = ::time(NULL) - this->running[i]->starttime;

          int speed = 0;

          if (tdiff > 0)
            speed = (this->running[i]->bytes / tdiff / 1024);
          
          overall_speed += speed;

          
          snprintf(tmp, 1024, "%30.30s @ %20.20s (%d%%, %d kB/s)\033[K\n",
                   this->running[i]->messageid,
                   this->running[i]->server,
                   percent,
                   speed);

          fprintf(stdout, "%s", tmp);
          lines++;
        }
    }

  char timeString[100];
  
  timeString[0] = '\0';

  if (overall_speed > 0)
    {
      long long remain_sec = this->queueSize / (overall_speed * 1024);
      int h = 0;
      int m = 0;
      int s = 0;
      while (remain_sec > 3600)
        {
          h++;
          remain_sec -= 3600;
        }

      while (remain_sec > 60)
        {
          m++;
          remain_sec -= 60;
        }

      s = remain_sec;

      sprintf(timeString, "~ %.2d:%.2d:%.2d", h, m, s);
    }

      

  snprintf(tmp, 1024, "%d kB/s, %d MB queued (%s)\033[K\n",
           overall_speed, (int) (this->queueSize / 1024 / 1024), timeString);
  fprintf(stdout, "%s", tmp);
  lines++;

  pthread_mutex_unlock(&this->mutex_running);

  // add message to go back some lines

  snprintf(tmp, 1024, "\r\033[%dA", lines);

  this->messages.push_back(strdup(tmp));

  pthread_mutex_unlock(&this->mutex_messages);

  fflush(stdout);
}
