/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.6 $
 * $Date: 2004/03/31 15:48:58 $
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>

#include "global.h"
#include "NNTPConnection.h"
#include "Connection.h"
#include "NewsServer.h"

NNTPConnection::NNTPConnection(NewsServer* server) : Connection(server)
{
  this->unavailableGroups.clear();
  this->activeGroup = NULL;
}

NNTPConnection::~NNTPConnection()
{
  for (unsigned int i = 0; i < this->unavailableGroups.size(); i++)
    {
      free(this->unavailableGroups[i]);
      this->unavailableGroups[i] = NULL;
    }

  this->unavailableGroups.clear();
}

char* NNTPConnection::request(char* req)
{
  if (!req)
    return NULL;

  this->write(req);
  char* answer = this->readLine();

  if (!answer)
    return NULL;

  if (!strncmp(answer, "480", 3))
    {
      info("%s requested authorization", this->server->getHost());

      //authentication required!
      if (this->authenticate() < 0)
        return NULL;
      
      //try again
      this->write(req);
      answer = this->readLine();
      return answer;
    }

  return answer;
}

int NNTPConnection::authenticate()
{
  char tmp[1024];

  if ((!((NewsServer*)this->server)->getUser()) || 
      (!((NewsServer*)this->server)->getPass()))
    return -1;
    
  
  snprintf(tmp, 1024, "AUTHINFO USER %s\r\n", ((NewsServer*)this->server)->getUser());

  this->write(tmp);

  char* answer = this->readLine();
  
  if (!answer)
    return -1;

  if (!strncmp(answer, "2", 1))
    {
      info("authorization for %s successful", this->server->getHost());
      return 0;
    }

  if (!strncmp(answer, "3", 1))
    {
      snprintf(tmp, 1024, "AUTHINFO PASS %s\r\n", ((NewsServer*)this->server)->getPass());
      
      this->write(tmp);

      answer = this->readLine();
      if (!answer)
        {
          error("authorization for %s failed: Connection closed by remote host.", this->server->getHost());
          return -1;
        }

      if (!strncmp(answer, "2", 1))
        {
          info("authorization for %s successful", this->server->getHost());
          return 0;
        }

      error("authorization for %s failed (Answer: %s)", this->server->getHost(), answer);
      return -1;
    }

  return -1;
}

int NNTPConnection::doConnect()
{
  info("Opening connection to %s", this->getServer()->getHost());
  int res = Connection::doConnect();
  if (res < 0)
    return res;
  char* answer = this->doReadLine();

  if (!answer)
    {
      error("Connection to %s failed: Connection closed by remote host.", this->server->getHost());
      return -1;
    }


  if (strncmp(answer, "2", 1))
    {
      error("Connection to %s failed. Answer: ", this->server->getHost(), answer);
      return -1;
    }

  info("Connection to %s established.", this->getServer()->getHost());
  return 0;
}

int NNTPConnection::doDisconnect()
{
  this->request("quit\r\n");
  return Connection::doDisconnect();
}


int NNTPConnection::joinGroup(char* grp)
{
  if (!grp)
    {
      debug("joinGroup called with NULL-pointer!!");
      return -1;
    }

  if ((this->activeGroup) && (!strcmp(this->activeGroup, grp)))
    return 0;
  
  for (unsigned int i = 0; i < this->unavailableGroups.size(); i++)
    {
      if (!strcmp(grp, this->unavailableGroups[i]))
        {
          debug("Group %s unavailable on %s.", grp, this->getServer()->getHost());
          return -1;
        }
    }

  char tmp[1024];
  snprintf(tmp, 1024, "GROUP %s\r\n", grp);

  char* answer = this->request(tmp);

  if ((answer) && (!strncmp(answer, "2", 1)))
    {
      info("Changed group to %s on %s", grp, this->getServer()->getHost());

      if (this->activeGroup)
        free(this->activeGroup);
      
      this->activeGroup = strdup(grp);
      return 0;
    }

  if (!answer)
    {
      warn("Error changing group on %s: Connection closed by remote host.",
           this->getServer()->getHost());
      return -1;
    }
  else
    {
      warn("Error changing group on %s to %s: Answer was \"%s\".",
           this->getServer()->getHost(), grp, answer);      
      this->unavailableGroups.push_back(strdup(grp));
    }

  return -1;
}
